# system modules
import json
import logging


logger = logging.getLogger(__name__)


def from_jsonlines(string):
    if hasattr(string, "decode"):
        string = string.decode(errors="ignore")
    string = str(string or "")
    for i, line in enumerate(string.splitlines(), start=1):
        try:
            yield json.loads(line)
        except (json.JSONDecodeError, ValueError, TypeError) as e:
            logger.warning(f"line #{i} ({line!r}) is invalid JSON: {e!r}")
            continue



# system modules
import time
import collections
import uuid
import os
import json
import re
import textwrap
import sys
import shlex
import logging
import subprocess
import argparse
import datetime
from datetime import datetime as dt
from pathlib import Path

# internal modules
from annextimelog.event import Event
from annextimelog.log import stdout, stderr
from annextimelog.run import run, get_repo_root
from annextimelog import utils

# external modules
import rich
from rich.logging import RichHandler
from rich.console import Console
from rich.syntax import Syntax
from rich.panel import Panel
from rich.columns import Columns
from rich.table import Table
from rich.pretty import Pretty
from rich.text import Text
from rich import box

logger = logging.getLogger(__name__)


def git_cmd_handler(args, other_args):
    result = run(subprocess.Popen, ["git", "-C", str(args.repo)] + other_args)
    result.wait()
    sys.exit(result.returncode)


def sync_cmd_handler(args, other_args):
    if logger.getEffectiveLevel() < logging.DEBUG:
        with run(
            subprocess.Popen, ["git", "-C", args.repo, "annex", "assist"]
        ) as process:
            process.wait()
            sys.exit(process.returncode)
    else:
        with stderr.status("Syncing..."):
            result = run(subprocess.run, ["git", "-C", args.repo, "annex", "assist"])
    if result.returncode or result.stderr:
        if result.returncode:
            logger.error(f"Syncing failed according to git annex.")
        if result.stderr:
            logger.warning(
                f"git annex returned some error messages. "
                f"This might be harmless, maybe try again (a couple of times)."
            )
        sys.exit(1)
    else:
        logger.info(f"✅ Syncing finished")
    sys.exit(result.returncode)


def track_cmd_handler(args, other_args):
    event = Event(repo=args.repo)
    for i, arg in enumerate(other_args, start=1):
        if t := Event.parse_date(arg):
            logger.debug(f"Interpreted argument #{i} {arg!r} as {Event.timeformat(t)}")
            if event.start is None:
                event.start = t
            elif event.end is None:
                event.end = t
            else:
                logger.warning(
                    f"Ignoring argument #{i} {arg!r}. "
                    f"Looks like a time but we already have two times {event.start = }, {event.end = }."
                )
            continue
        if m := re.search(r"^@(.*)$", arg.strip()):
            location = m.group(1).strip()
            logger.debug(f"argument #{i} {arg!r} means {location = !r}")
            event.location.add(location)
            continue
        if m := re.search(r"^:(.*)$", arg.strip()):
            note = m.group(1).strip()
            logger.debug(f"argument #{i} {arg!r} is a note {note = !r}")
            if event.note:
                logger.warning(f"Overwriting previous note {note!r}")
            event.note = note
            continue
        if m := re.search(r"^=(.*)$", arg.strip()):
            title = m.group(1).strip()
            logger.debug(f"argument #{i} {arg!r} is a {title = !r}")
            if event.title:
                logger.warning(f"Overwriting previous title {title!r}")
            event.title = title
            continue
        if m := re.search(r"(?P<field>\S+?)(?P<operator>[+]?)=(?P<value>.*)", arg):
            field, operator, value = m.groups()
            if operator.startswith("+"):
                logger.debug(
                    f"argument #{i} {arg!r} adds {value!r} to metadata field {field!r}"
                )
                event.fields[field].add(value)
            else:
                logger.debug(
                    f"argument #{i} {arg!r} sets metadata field {field!r} to (only) {value!r}"
                )
                event.fields[field] = {value}
            continue
        logger.debug(f"argument #{i} {arg!r} is a tag")
        event.tags.add(arg)
    if not (event.start and event.end):
        logger.critical(
            f"Currently, 'annextimelog track' can only record events with exactly two given time bounds. "
            f"This will change in the future."
        )
        sys.exit(1)
    if logger.getEffectiveLevel() < logging.DEBUG:
        logger.debug(f"Event before saving:")
        stderr.print(event.to_rich())
    # make the file
    event.store()
    cmd = ["git", "-C", args.repo, "annex", "metadata", "--key", event.key]
    cmd.extend(["--set", f"start={event.timeformat(event.start)}"])
    cmd.extend(["--set", f"end={event.timeformat(event.end)}"])
    if event.title:
        cmd.extend(["--set", f"title={event.title}"])
    if event.note:
        cmd.extend(["--set", f"note={event.note}"])
    for tag in event.tags:
        cmd.extend(["--tag", tag])
    for location in event.location:
        cmd.extend(["--set", f"location+={location}"])
    for field, values in event.fields.items():
        for value in values:
            cmd.extend(["--set", f"{field}+={value}"])
    if run(subprocess.run, cmd).returncode:
        logger.error(
            f"Something went wrong setting annex metadata on event {event.id} key {event.key!r}"
        )
    if logger.getEffectiveLevel() <= logging.DEBUG and args.output_format not in (
        "rich",
        "console",
    ):
        logger.debug(f"Event after saving:")
        stderr.print(event.to_rich())
    # output event
    stdout.print(getattr(event, f"to_{args.output_format}", event.to_rich)())


def summary_cmd_handler(args, other_args):
    if other_args:
        logger.warning(f"Ignoring other arguments {other_args}")
    with logger.console.status(f"Querying metadata..."):
        result = run(
            subprocess.run, ["git", "-C", args.repo, "annex", "metadata", "--json"]
        )
    infos = dict()
    paths = collections.defaultdict(set)
    for info in utils.from_jsonlines(result.stdout):
        path = Path(info.get("input", [None])[0])
        infos[(info.get("key"), path.stem)] = info.get("fields", dict())
        paths[info.get("key")].add(path)
    for (key, id), fields in infos.items():
        event = Event(
            repo=args.repo,
            id=id,
            key=key,
            fields={
                k: set(v)
                for k, v in fields.items()
                if not (k.endswith("-lastchanged") or k in ["lastchanged"])
            },
        )
        logger.debug(f"{event = }")
        stdout.print(getattr(event, f"to_{args.output_format}", event.to_rich)())
    if not infos:
        logger.info(
            f"😴 There are no tracked events yet. Add some with 'annextimelog track ...'"
        )


parser = argparse.ArgumentParser(
    description="⏱️ Time tracker based on Git Annex",
    prog="annextimelog",
    formatter_class=argparse.RawDescriptionHelpFormatter,
)

datagroup = parser.add_argument_group(title="Data")
datagroup.add_argument(
    "--repo",
    type=Path,
    default=(
        default := Path(
            os.environ.get("ANNEXTIMELOGREPO")
            or Path(os.environ.get("XDG_DATA_HOME", Path.home() / ".local" / "share"))
            / "annextimelog"
        )
    ),
    help=f"Backend repository to use. "
    f"Defaults to $ANNEXTIMELOG_REPO or $XDG_DATA_HOME/annextimelog ({str(default)})",
)

parser.add_argument(
    "--force",
    action="store_true",
    help="Just do it. Ignore potential data loss.",
)

outputgroup = parser.add_argument_group(
    title="Output", description="Options changing output behaviour"
)
outputgroup.add_argument(
    "-v",
    "--verbose",
    action="count",
    default=0,
    help="verbose output. More -v ⮕ more output",
)
outputgroup.add_argument(
    "-q",
    "--quiet",
    action="count",
    default=0,
    help="less output. More -q ⮕ less output",
)
outputgroup.add_argument(
    "-O",
    "--output-format",
    choices={"rich", "console", "json", "timeclock"},
    default=(default := "console"),
    help="Select output format. Defaults to {default!r}.",
)


subparsers = parser.add_subparsers(title="Subcommands", dest="subcommand")
gitparser = subparsers.add_parser(
    "git", help="Access the underlying git repository", add_help=False
)
gitparser.set_defaults(func=git_cmd_handler)
syncparser = subparsers.add_parser(
    "sync",
    help="sync data",
    description=textwrap.dedent(
        """
    Sync data with configured remotes.
    """
    ).strip(),
    aliases=["sy"],
)
syncparser.set_defaults(func=sync_cmd_handler)
trackparser = subparsers.add_parser(
    "track",
    help="record a time period",
    description=textwrap.dedent(
        """
    Record a time period.

    """
    ).strip(),
    aliases=["tr"],
)
trackparser.set_defaults(func=track_cmd_handler)
summaryparser = subparsers.add_parser(
    "summary",
    help="show a summary of tracked periods",
    description=textwrap.dedent(
        """
    Show a summary of tracked periods

    """
    ).strip(),
    aliases=["su"],
)
summaryparser.set_defaults(func=summary_cmd_handler)


def cli(args=None):
    args, other_args = parser.parse_known_args(args=args)

    logging.basicConfig(
        level=(level := logging.INFO - (args.verbose - args.quiet) * 5),
        format="%(message)s",
        datefmt="[%X]",
        handlers=[
            RichHandler(
                console=stderr,  # log to stderr
                rich_tracebacks=True,
                show_path=level < logging.DEBUG - 10,
            )
        ],
    )
    logger.debug(f"{args = }")
    logger.debug(f"{other_args = }")

    if args.repo.exists() and not args.repo.is_dir():
        logger.critical(f"{args.repo} exists but is not a directory.")
        sys.exit(1)

    if args.repo.exists():
        logger.debug(f"{args.repo} exists")
        if repo_root := get_repo_root(args.repo):
            if repo_root.resolve() != args.repo.resolve():
                logger.critical(
                    f"There's something funny with {args.repo}: git says the repo root is {repo_root}. "
                )
                sys.exit(1)
        else:
            logger.critical(f"{args.repo} exists but is no git repository. 🤔")
            sys.exit(1)
    else:
        if not args.repo.parent.exists():
            logger.info(f"📁 Creating {args.repo.parent}")
            args.repo.parent.mkdir(parent=True, exist_ok=True)
        logger.info(f"Making a git repository at {args.repo}")
        result = run(
            subprocess.run, ["git", "init", str(args.repo)], capture_output=True
        )
        if result.returncode:
            logger.error(f"Couldn't make git repository at {args.repo}")
            sys.exit(1)

    # ✅ at this point, args.repo is a git repository
    annex_uuid = None
    logger.debug(f"Making sure {args.repo} is a git annex repository")
    if not (
        result := run(
            subprocess.run,
            ["git", "-C", args.repo, "config", "annex.uuid"],
            title=f"get git annex uuid for {args.repo}",
            debug_on_error=False,
        )
    ).returncode:
        annex_uuid = result.stdout.strip()
    if not annex_uuid:
        logger.debug(f"{args.repo} is not a git annex repository")
        if not (
            result := run(
                subprocess.run,
                ["git", "-C", args.repo, "annex", "init"],
                title=f"add an annex to {args.repo}",
            )
        ).returncode:
            logger.info(f"Added an annex to {args.repo}")
        else:
            logger.critical(f"Couldn't add an annex to {args.repo}")
            sys.exit(1)

    # ✅ at this point, args.repo is a git annex repository

    result = run(subprocess.run, ["git", "-C", args.repo, "status", "--porcelain"])
    if result.returncode or result.stdout or result.stderr:
        logger.warning(
            f"🐛 The repo {args.repo} is not clean. "
            f"This should not happen. Committing away the following changes:"
        )
        result = run(subprocess.Popen, ["git", "-C", args.repo, "status"])
        with logger.console.status("Committing..."):
            result = run(subprocess.run, ["git", "-C", args.repo, "annex", "add"])
            result = run(subprocess.run, ["git", "-C", args.repo, "add", "-A"])
            result = run(
                subprocess.run,
                ["git", "-C", args.repo, "commit", "-m", "🤷 Leftover changes"],
            )
        result = run(subprocess.run, ["git", "-C", args.repo, "status", "--porcelain"])
        if not (result.returncode or result.stderr):
            logger.info(f"✅ Repo is now clean")
        else:
            logger.warning(f"Commiting leftover changes didn't work.")

    # handle the subcommand
    # (when a subcommand is specified, the 'func' default is set to a callback function)
    if not getattr(args, "func", None):
        # default to 'atl summary'
        args.func = summary_cmd_handler
    try:
        args.func(args, other_args)
    finally:
        result = run(subprocess.run, ["git", "-C", args.repo, "status", "--porcelain"])
        if result.returncode or result.stdout or result.stderr:
            logger.warning(
                f"🐛 This command left the repo {args.repo} in an unclean state. "
                f"This should not happen. Consider investigating. "
                f"The next time you run any 'annextimelog' command, these changes will be committed."
            )
            result = run(subprocess.Popen, ["git", "-C", args.repo, "status"])


if __name__ == "__main__":
    cli()



# system modules
import logging
import locale

# external modules
from rich.console import Console

stdout = Console()
stderr = Console(stderr=True)

# Allow locale for strftime etc.
locale.setlocale(locale.LC_ALL, "")


def logger_console_getter(logger):
    for handler in logger.root.handlers:
        if console := getattr(handler, "console", None):
            logger._console = console
            return console
    return Console(stderr=True)  # last resort: make temporary new one


# monkey-patch Logger to have a 'console' property returning the actually used
# console in the RichHandler
logging.Logger.console = property(logger_console_getter)



# system modules
import re
import subprocess
import locale
import logging
import textwrap
import collections
import string
import random
import datetime
from datetime import datetime as dt
from pathlib import Path
from dataclasses import dataclass, asdict
from typing import Optional, Set, Dict

# internal modules
from annextimelog.run import run
from annextimelog import utils

# external modules
from rich.table import Table
from rich.text import Text
from rich.highlighter import ReprHighlighter, ISO8601Highlighter
from rich import box

logger = logging.getLogger(__name__)


@dataclass
class Event:
    repo: Path
    id: Optional[Path] = None
    paths: Optional[Set[Path]] = None
    key: Optional[str] = None
    fields: Optional[Dict[str, Set[str]]] = None

    SUFFIX = ".ev"

    def __post_init__(self):
        if self.id is None:
            self.id = self.random_id()
        if self.paths is None:
            self.paths = set()
        if self.fields is None:
            self.fields = collections.defaultdict(set)

    @property
    def location(self):
        if "location" not in self.fields:
            self.fields["location"] = set()
        return self.fields["location"]

    @property
    def start(self):
        if not "start" in self.fields:
            self.fields["start"] = set()
        if not (start := self.fields["start"]):
            return None
        elif len(start) > 1:
            try:
                earliest = min(filter(bool, map(self.parse_date, start)))
            except Exception as e:
                logger.error(
                    f"There are len({start}) start times for event {self!r}, but I can't determine the earliest."
                )
                self.fields["start"].clear()
                return None
            self.fields["start"].clear()
            self.fields["start"].add(earliest)
        return self.parse_date(next(iter(self.fields["start"]), None))

    @start.setter
    def start(self, value):
        if value is None:
            self.fields["start"].clear()
            return
        if d := self.parse_date(value):
            self.fields["start"].clear()
            self.fields["start"].add(d)
        else:
            logger.error(f"Couldn't interpret {value!r} as time.")
            self.fields["start"].clear()

    @property
    def end(self):
        if not "end" in self.fields:
            self.fields["end"] = set()
        if not (end := self.fields["end"]):
            return None
        elif len(end) > 1:
            try:
                latest = max(filter(bool, map(self.parse_date, end)))
            except Exception as e:
                logger.error(
                    f"There are len({end}) end times for event {self!r}, but I can't determine the earliest."
                )
                self.fields["end"].clear()
                return None
            self.fields["end"].clear()
            self.fields["end"].add(latest)
        return self.parse_date(next(iter(self.fields["end"]), None))

    @end.setter
    def end(self, value):
        if value is None:
            self.fields["end"].clear()
            return
        if d := self.parse_date(value):
            self.fields["end"].clear()
            self.fields["end"].add(d)
        else:
            logger.error(f"Couldn't interpret {value!r} as time.")
            self.fields["end"].clear()

    @property
    def note(self):
        if len(note := self.fields.get("note", set())) > 1:
            note = "\n".join(self.fields["note"])
            self.fields["note"].clear()
            self.fields["note"].add(note)
        return "\n".join(self.fields.get("note", set()))

    @note.setter
    def note(self, value):
        self.fields["note"].clear()
        self.fields["note"].add(value)

    @property
    def title(self):
        if len(title := self.fields.get("title", set())) > 1 or any(
            re.search(r"[\r\n]", t) for t in title
        ):
            title = " ".join(re.sub(r"[\r\n]+", " ", t) for t in self.fields["title"])
            self.fields["title"].clear()
            self.fields["title"].add(title)
        return "\n".join(self.fields.get("title", set()))

    @title.setter
    def title(self, value):
        value = re.sub(r"[\r\n]+", " ", str(value))
        self.fields["title"].clear()
        self.fields["title"].add(value)

    @property
    def tags(self):
        if "tag" not in self.fields:
            self.fields["tag"] = set()
        return self.fields["tag"]

    @classmethod
    def from_fields(cls, fields):
        event = cls()

    @staticmethod
    def random_id():
        return "".join(random.choices(string.ascii_letters + string.digits, k=8))

    @staticmethod
    def parse_date(string):
        if isinstance((d := string), datetime.datetime):
            if not d.tzinfo:
                d = d.replace(tzinfo=datetime.timezone.utc)
            return d
        offset = datetime.timedelta(days=0)
        if m := re.search(r"^(?P<prefix>[yt]+)(?P<rest>.*)$", string):
            offset = datetime.timedelta(
                days=sum(dict(y=-1, t=1).get(c) for c in m.group("prefix"))
            )
            logger.debug(
                f"{string!r} starts with {m.group('prefix')!r}, so thats as an {offset = }"
            )
            string = m.group("rest")
        result = None
        for f in (
            dt.fromisoformat,
            lambda s: dt.fromisoformat(f"{dt.utcnow().strftime('%Y-%m-%d')} {s}"),
            lambda s: dt.strptime(
                f"{dt.utcnow().strftime('%Y-%m-%d')} {s}", "%Y-%m-%d %H%M"
            ),
            lambda s: dt.strptime(
                f"{dt.utcnow().strftime('%Y-%m-%d')} {s}", "%Y-%m-%d %H"
            ),
            lambda s: dt.strptime(
                f"{dt.utcnow().strftime('%Y-%m-%d')} {s}", "%Y-%m-%d %-H"
            ),
        ):
            try:
                result = f(string).replace(tzinfo=datetime.timezone.utc)
                break
            except Exception:
                pass
        if result:
            result += offset
        return result

    @staticmethod
    def timeformat(t):
        return t.strftime("%Y-%m-%dT%H:%M:%S%z")

    def store(self):
        self.start = self.start or datetime.date.utcnow()
        self.end = self.end or datetime.date.utcnow()
        if self.end < self.start:
            logger.info(
                f"↔️  event {self.id!r}: Swapping start and end (they're backwards)"
            )
            self.start, self.end = self.end, self.start

        def folders():
            start, end = self.start, self.end
            start = datetime.date(start.year, start.month, start.day)
            end = datetime.date(end.year, end.month, end.day)
            day = start
            lastweekpath = None
            while day <= end:
                path = Path()
                for p in "%Y %m %d".split():
                    path /= day.strftime(p)
                yield path
                weekpath = Path()
                for p in "%Y W %W".split():
                    weekpath /= day.strftime(p)
                if weekpath != lastweekpath:
                    yield weekpath
                lastweekpath = weekpath
                day += datetime.timedelta(days=1)

        paths = set()
        for folder in folders():
            if not (folder_ := self.repo / folder).exists():
                logger.debug(f"📁 Creating new folder {folder}")
                folder_.mkdir(parents=True)
            file = (folder_ / self.id).with_suffix(self.SUFFIX)
            if (file.exists() or file.is_symlink()) and not (self.paths or self.key):
                logger.warning(
                    f"🐛 {file} exists although this event {event.id} is new (it has no paths or key attached). "
                    f"This is either a bug 🐛 or you just witnessed a collision. 💥"
                    f"🗑️ Removing {file}."
                )
                file.unlink()
            if file.is_symlink() and not os.access(str(file), os.W_OK):
                logger.debug(f"🗑️ Removing existing read-only symlink {file}")
                file.unlink()
            file_existed = file.exists()
            with file.open("w") as fh:
                logger.debug(
                    f"🧾 {'Overwriting' if file_existed else 'Creating'} {file} with content {self.id!r}"
                )
                fh.write(self.id)
            try:
                paths.add(file.relative_to(self.repo))
            except ValueError:
                paths.add(file)
        self.paths = paths
        with logger.console.status(f"Adding {len(self.paths)} paths..."):
            result = run(
                subprocess.run,
                ["git", "-C", self.repo, "annex", "add", "--json"] + sorted(self.paths),
                output_lexer="json",
                title=f"Adding {len(self.paths)} paths for event {self.id!r}",
            )
            keys = set()
            for info in utils.from_jsonlines(result.stdout):
                if key := info.get("key"):
                    keys.add(key)
            if len(keys) != 1:
                logger.warning(
                    f"🐛 Adding {len(self.paths)} paths for event {self.id!r} resulted in {len(keys)} keys {keys}. "
                    f"That should be exactly 1. This is probably a bug."
                )
            if keys:
                self.key = next(iter(keys), None)
                logger.debug(f"🔑 key for event {self.id!r} is {self.key!r}")
        with logger.console.status(f"Force-dropping {keys = }..."):
            result = run(
                subprocess.run,
                ["git", "-C", self.repo, "annex", "drop", "--force", "--key"]
                + list(keys),
                title=f"Force-dropping {keys = } for event {self.id!r}",
            )
        with logger.console.status(f"Committing addition of event {self.id!r}..."):
            result = run(
                subprocess.run,
                [
                    "git",
                    "-C",
                    self.repo,
                    "commit",
                    "-m",
                    f"➕ Add {self.id!r} ({self.title or 'untitled'})",
                ],
                title=f"Committing addition of event {self.id!r}",
            )
            if not result.returncode:
                logger.info(f"✅ Committed addition of event {self.id!r}")

    def to_rich(self):
        table = Table(title=self.title, padding=0, box=box.ROUNDED, show_header=False)
        table.add_column("", justify="left")
        table.add_column("Field", justify="right", style="cyan")
        table.add_column("Value", justify="left")
        if self.id:
            table.add_row("💳", "id", f"[b]{self.id}[/b]")
        if self.paths and logger.getEffectiveLevel() < logging.DEBUG:
            table.add_row(
                "🧾",
                "paths",
                ReprHighlighter()(Text("\n".join(str(p) for p in self.paths))),
            )
        if self.paths and logger.getEffectiveLevel() < logging.DEBUG:
            table.add_row("🔑", "key", self.key)
        timehighlighter = ISO8601Highlighter()
        if start := self.start:
            table.add_row("🚀", "start", start.strftime("%c%Z"))
        if end := self.end:
            table.add_row("⏱️", "end", end.strftime("%c%Z"))
        if self.location:
            table.add_row(
                "📍", "location", ", ".join([f"📍 {t}" for t in sorted(self.location)])
            )
        if self.tags:
            table.add_row(
                "🏷️", "tags", " ".join([f"🏷️ {t}" for t in sorted(self.tags)])
            )
        for field, values in self.fields.items():
            if field in "start end tag location title note".split():
                continue
            table.add_row("", field, " ".join(f"📝 {value}" for value in values))
        if self.note:
            table.add_row("📝", "note", self.note)
        return table

    def to_dict(self):
        return asdict(self)

    def to_json(self):
        def default(x):
            if hasattr(x, "strftime"):
                return self.timeformat(x)
            if not isinstance(x, str):
                try:
                    iter(x)
                    return tuple(x)
                except TypeError:
                    pass
            return str(x)

        return json.dumps(self.to_dict(), default=default)

    def to_timeclock(self):
        def sanitize(s):
            s = re.sub(r"[,:;]", r"⁏", s)  # replace separation chars
            s = re.sub(r"[\r\n]+", r" ", s)  # no newlines
            return s

        hledger_tags = {
            "tags": " ⁏ ".join(map(sanitize, sorted(self.tags))),
            "location": " ⁏ ".join(map(sanitize, sorted(self.location))),
        }
        for tag in sorted(self.tags):
            hledger_tags[tag] = ""
        hledger_tags = [f"{t}: {v}" for t, v in hledger_tags.items()]
        hledger_comment = f";  {', '.join(hledger_tags)}" if hledger_tags else ""
        info = [
            ":".join(self.fields.get("account", self.tags)),
            self.title,
            hledger_comment,
        ]
        return textwrap.dedent(
            f""" 
        i {self.start.strftime('%Y-%m-%d %H:%M:%S%z')} {'  '.join(filter(bool,info))}
        o {self.end.strftime('%Y-%m-%d %H:%M:%S%z')}
        """
        ).strip()
